#include <pth.h>
#include <stdlib.h>
#include <unistd.h>
#include "constants.h"
#include "device.h"
#include "dumb_resample.h"
#include "mixer.h"
#include "output.h"
#include "output_stream.h"
#include "unix_listen.h"
#include "unix_serve.h"

#define BUFFER_SIZE 1024

/* Based on the resampler class in Audiere (http://audiere.sf.net/). */
static inline short
_u8tos16 (unsigned char u)
{
  return ((short)u - 128) * 256;
}

static inline int
_clamp (int min, int val, int max)
{
  if (val < min)
    return min;
  else if (val > max)
    return max;
  else
    return val;
}

/* Also based on the Resampler class in Audiere. */
static int
_fill_buffers (int sock, int sample_bytes, int channels, int rate,
	       sample_t *out_l, sample_t *out_r, int *buffer_length)
{
  unsigned char initial_buffer[BUFFER_SIZE * 4];
  int frame_size = sample_bytes * channels;
  unsigned read = pth_read (sock, initial_buffer, BUFFER_SIZE * frame_size);
  int bytes_received;

  if (read < 0)
    read = 0;

  bytes_received = read;
  read /= frame_size;

  if (channels == 1)
    {
      if (sample_bytes == 1)
	{
	  unsigned char *in = initial_buffer;
	  unsigned i;

	  for (i = 0; i < read; i++)
	    {
	      short sample = _u8tos16 (*in++);
	      *out_l++ = sample;
	    }
	}
      else
	{
	  short *in = (short *) initial_buffer;
	  unsigned i;

	  for (i = 0; i < read; i++)
	    {
	      short sample = *in++;
	      *out_l++ = sample;
	    }
	}
    }
  else
    {
      if (sample_bytes == 1)
	{
	  unsigned char *in = initial_buffer;
	  unsigned i;

	  for (i = 0; i < read; i++)
	    {
	      *out_l++ = _u8tos16 (*in++);
	      *out_r++ = _u8tos16 (*in++);
	    }
	}
      else
	{
	  short *in = (short *) initial_buffer;
	  unsigned i;

	  for (i = 0; i < read; i++)
	    {
	      *out_l++ = *in++;
	      *out_r++ = *in++;
	    }
	}
    }

  *buffer_length = read;
  return bytes_received;
}

static void
_handler (int sock, void *data)
{
  struct output_stream *s = NULL;
  sample_t *buffer_l = NULL;
  sample_t *buffer_r = NULL;
  sample_t *tmp_l = NULL;
  sample_t *tmp_r = NULL;
  DUMB_RESAMPLER resampler_l;
  DUMB_RESAMPLER resampler_r;
  short *out = NULL;
  int index = -1, tmp, orig_sample_bytes = -1, orig_channels = -1,
    orig_rate = -1, buffer_length = 0;
  float delta = 1.0;

  pth_read (sock, &index, sizeof (index));

  if (index >= 0)
    s = mixer_get_stream (index);

  if (s == NULL || s->done_writing)
    return;

  output_stream_ref (s);
  pth_read (sock, &tmp, sizeof (tmp));
  orig_sample_bytes = tmp & 0xffff;
  orig_channels = tmp >> 16;
  pth_read (sock, &orig_rate, sizeof (orig_rate));
  delta = orig_rate / (float) device_get_rate ();
  s->format_ratio = delta * orig_sample_bytes / device_get_sample_bytes () *
    orig_channels / device_get_channels ();
  buffer_l = (sample_t *) malloc (sizeof (sample_t) * BUFFER_SIZE);
  buffer_r = (sample_t *) malloc (sizeof (sample_t) * BUFFER_SIZE);
  tmp_l = (sample_t *) malloc (sizeof (sample_t) * BUFFER_SIZE);
  tmp_r = (sample_t *) malloc (sizeof (sample_t) * BUFFER_SIZE);
  out = (short *) malloc (BUFFER_SIZE * 4);
  s->bytes_received +=
    _fill_buffers (sock, orig_sample_bytes, orig_channels, orig_rate, buffer_l,
		   buffer_r, &buffer_length);
  dumb_reset_resampler (&resampler_l, buffer_l, 0, 0, buffer_length);

  if (orig_channels == 2)
    dumb_reset_resampler (&resampler_r, buffer_r, 0, 0, buffer_length);

  while (!s->closing)
    {
      int rv, i;
      short *p = out;
      memset (tmp_l, 0, BUFFER_SIZE * sizeof (sample_t));
      rv = dumb_resample (&resampler_l, tmp_l, BUFFER_SIZE, 1.0, delta);

      if (rv == 0)
	{
	  s->bytes_received +=
	    _fill_buffers (sock, orig_sample_bytes, orig_channels, orig_rate,
			   buffer_l, buffer_r, &buffer_length);

	  if (buffer_length == 0)
	    break;
	  else
	    {
	      resampler_l.pos = resampler_l.subpos = resampler_l.start = 0;
	      resampler_l.end = buffer_length;
	      resampler_l.dir = 1;
	      resampler_r.pos = resampler_r.subpos = resampler_r.start = 0;
	      resampler_r.end = buffer_length;
	      resampler_r.dir = 1;
	      continue;
	    }
	}

      if (orig_channels == 2)
	{
	  int rv2;
	  memset (tmp_r, 0, BUFFER_SIZE * sizeof (sample_t));
	  rv2 = dumb_resample (&resampler_r, tmp_r, BUFFER_SIZE, 1.0,
			       delta);

	  for (i = 0; i < rv; i++)
	    {
	      *p++ = _clamp (-32768, tmp_l[i], 32767);
	      *p++ = _clamp (-32768, tmp_r[i], 32767);
	    }
	}
      else
	{
	  for (i = 0; i < rv; i++)
	    {
	      short sample = _clamp (-32768, tmp_l[i], 32767);
	      *p++ = sample;
	      *p++ = sample;
	    }
	}

      output_stream_write (s, out, rv * 4);
    }

  s->done_writing = TRUE;
  free (tmp_r);
  free (tmp_l);
  free (out);
  free (buffer_r);
  free (buffer_l);
  output_stream_unref (s);
  s = NULL;
}

static void *
_server_thread (void *_arg)
{
  int sock = unix_listen ("/dev/usound/out", 3);
  unix_serve (sock, _handler, NULL);
  return NULL;
}

void
output_start (void)
{
  pth_spawn (PTH_ATTR_DEFAULT, _server_thread, NULL);
}
