#include <pth.h>
#include <stdio.h>
#include <stdlib.h>
#include "constants.h"
#include "device.h"
#include "output_stream.h"
#include "ringbuffer.h"

struct output_stream *
output_stream_create (void)
{
  int bufsize = device_get_fragment_size () * OUTPUT_STREAM_FRAGMENTS;
  void *buffer = malloc (bufsize);
  struct output_stream *s =
    (struct output_stream *) malloc (sizeof (struct output_stream));
  RingBuffer_Init (&s->rbuf, bufsize, buffer);
  pth_mutex_init (&s->mutex);
  pth_cond_init (&s->cond);
  pth_cond_init (&s->sync_cond);
  s->volume = 65536;
  s->playing = FALSE;
  s->ref_count = 0;
  s->resetting = FALSE;
  s->closing = FALSE;
  s->bytes_received = 0;
  s->bytes_played = 0;
  s->done_writing = FALSE;
  s->format_ratio = 1.0;
  return s;
}

void
output_stream_destroy (struct output_stream *s)
{
  free (s->rbuf.buffer);
  free (s);
}

int
output_stream_read (struct output_stream *s, void *buf, int bytes)
{
  int bytes_read;
  pth_mutex_acquire (&s->mutex, FALSE, NULL);
  bytes_read = RingBuffer_Read (&s->rbuf, buf, bytes);
  pth_cond_notify (&s->cond, FALSE);
  pth_mutex_release (&s->mutex);
  return bytes_read;
}

void
output_stream_write (struct output_stream *s, void *buf, int bytes)
{
  if (s->closing)
    return;

  pth_mutex_acquire (&s->mutex, FALSE, NULL);
  s->resetting = FALSE;

  while (bytes > 0 && !s->resetting)
    {
      int bytes_written = RingBuffer_Write (&s->rbuf, buf, bytes);
      bytes -= bytes_written;
      buf += bytes_written;

      if (bytes > 0)
	pth_cond_await (&s->cond, &s->mutex, NULL);
    }

  pth_mutex_release (&s->mutex);
}

void
output_stream_sync (struct output_stream *s)
{
  if (s->playing)
    {
      pth_mutex_acquire (&s->mutex, FALSE, NULL);
      pth_cond_await (&s->sync_cond, &s->mutex, NULL);
      pth_mutex_release (&s->mutex);
    }
}

void
output_stream_sync_notify (struct output_stream *s)
{
  pth_mutex_acquire (&s->mutex, FALSE, NULL);
  pth_cond_notify (&s->sync_cond, TRUE);
  pth_mutex_release (&s->mutex);
}

void
output_stream_ref (struct output_stream *s)
{
  s->ref_count++;
}

void
output_stream_unref (struct output_stream *s)
{
  if (--s->ref_count == 0)
    output_stream_destroy (s);
}

void
output_stream_reset (struct output_stream *s)
{
  pth_mutex_acquire (&s->mutex, FALSE, NULL);
  s->resetting = TRUE;
  RingBuffer_Flush (&s->rbuf);
  s->bytes_received = 0;
  s->bytes_played = 0;
  pth_cond_notify (&s->cond, FALSE);
  pth_mutex_release (&s->mutex);
}

void
output_stream_close (struct output_stream *s)
{
  s->closing = TRUE;
  output_stream_reset (s);
}

void
output_stream_set_buffer_size (struct output_stream *s, int buffer_size)
{
  pth_mutex_acquire (&s->mutex, FALSE, NULL);
  s->resetting = TRUE;
  RingBuffer_Flush (&s->rbuf);
  s->bytes_received = 0;
  s->bytes_played = 0;
  free (s->rbuf.buffer);
  void *buffer = malloc (buffer_size);
  RingBuffer_Init (&s->rbuf, buffer_size, buffer);
  pth_cond_notify (&s->cond, FALSE);
  pth_mutex_release (&s->mutex);
}
